import operator
from numba import types, njit, u1,u2,u4,u8, i8, carray
from numba.types import Tuple
from numba.typed import List
from numba.extending import overload_method, overload
from numba.experimental.structref import new
from cre.structref import define_structref_template
from llvmlite import ir
import numpy as np



vector_fields = [
    ("head", i8),
    ("data", i8[::1])
]

Vector, VectorTypeClass = define_structref_template("Vector",vector_fields)
VectorType = VectorTypeClass(fields=vector_fields)

@njit(cache=True)
def new_vector(size):
    '''
    Makes a new vector of size.
    '''
    st = new(VectorType)
    st.head = 0
    st.data = np.zeros(size,dtype=np.int64)
    return st

@njit(cache=True)
def _expand_to(self,size):
    if(len(self.data) < size):
        new_data = np.empty(size, dtype=np.int64)
        new_data[:len(self.data)] = self.data
        new_data[len(self.data):] = 0
        self.data = new_data

@overload_method(VectorTypeClass, "assert_size")
def assert_size(self,size):
    def impl(self,size):
        _expand_to(self,size)
    return impl

@njit(inline='never')
def _expand(self):
    '''
    Doubles the size of the vector. For whatever reason it runs faster when this
    is a separate function. Probably prevents some lookahead execution gone wrong.
    '''
    new_data = np.zeros(len(self.data)*2, dtype=np.int64)
    new_data[:len(self.data)] = self.data
    new_data[len(self.data):] = 0
    self.data = new_data

@overload_method(VectorTypeClass, "expand")
def assert_size(self):
    def impl(self):
        _expand(self)
    return impl

@overload_method(VectorTypeClass, "set_item_safe")
def vector_set_item_safe(self, i, x):
    ''' 
    Set's slot i to x. Ensures that data is large enough.
    '''
    def impl(self, i, x):
        if(i >= self.head): self.head = i+1
        _expand_to(self,self.head)
        self.data[i] = x
    return impl

@overload_method(VectorTypeClass, "add")
def vector_add(self, x):
    ''' 
    Adds an item to the vector and increments the head.
    '''
    def impl(self, x):
        if(self.head >= len(self.data)): _expand(self)
        self.data[self.head] = x
        self.head += 1
    return impl

@overload_method(VectorTypeClass, "pop")
def vector_pop(self):
    ''' 
    Returns the last item from the end of the vector and decrements the head.
    Returns zero if the vector is empty (does not raise an error).
    '''
    def impl(self):
        if(not self.head): return 0
        self.head -= 1
        return self.data[self.head]
    return impl



@overload_method(VectorTypeClass, "clear")
def vector_clear(self):
    ''' 
    Moves the head to zero, essentially clearing the vector.
    '''
    def impl(self):
        self.head = 0
    return impl

@overload_method(VectorTypeClass, "copy")
def vector_copy(self):
    ''' 
    Moves the head to zero, essentially clearing the vector.
    '''
    def impl(self):
        st = new(VectorType)
        st.head = self.head
        st.data = self.data.copy()
    return impl

@overload(operator.getitem)
def impl_getitem(self, i):
    if not isinstance(self, VectorTypeClass):
        return
    def impl(self,i):
        return self.data[i]
    return impl

@overload(operator.setitem)
def impl_setitem(self, i, x):
    if not isinstance(self, VectorTypeClass):
        return
    def impl(self,i,x):
        self.data[i] = x
    return impl

@overload(len)
def impl_len(self):
    if not isinstance(self, VectorTypeClass):
        return
    def impl(self):
        return self.head
    return impl






import timeit
N=100
def time_ms(f):
    f() #warm start
    return " %0.6f ms" % (1000.0*(timeit.timeit(f, number=N)/float(N)))



@njit
def test_vector():
    v = new_vector(10)
    for i in range(100):
        v.add(i)
        v[0] = i
        print(v[0],v[i],v.head,len(v.data))
    for i in range(50):
        print(v.pop())
    v.clear()
    print("H",v.head)
    print("L",len(v))

# test_vector()

@njit
def add_pop():
    v = new_vector(10000)
    for i in range(10000):
        v.add(i)
    for i in range(10000):
        v.pop()

@njit
def expand(self):
    # new_size = self.size*2
    new_data = np.empty(len(self.data)*2, dtype=np.int64)
    new_data[:len(self.data)] = self.data
    new_data[len(self.data):] = 0
    self.data = new_data
    # self.size = new_size

@njit
def dd(self, x):
    if(self.head >= len(self.data)): expand(self)
    self.data[self.head] = x
    self.head += 1

@njit
def add_pop_dd():
    v = new_vector(10000)
    for i in range(10000):
        dd(v,i)
    for i in range(10000):
        v.pop()


@njit
def add_pop_manual():
    v = new_vector(10000)
    for i in range(10000):
        v.data[i] = i
        v.head += i
        # v.add(i)
    for i in range(10000):
        v.pop()


@njit
def add_pop_inline():
    v = np.empty((10000,),dtype=np.int64)
    h = 0
    d = 0
    for i in range(10000):
        v[i] = i*i
        h += 1
    for i in range(10000):
        h -= 1
        d = v[h]





@njit
def v_new(size):
    v = np.empty((size+1,),dtype=np.int64)
    v[0] = 1
    return v

@njit
def v_expand(v):
    n_v = np.empty(len(v)*2, dtype=np.int64)
    n_v[:len(v)] = v
    return n_v

@njit
def v_add(v,x):
    if(v[0] > len(v)): v = v_expand(v)
    v[v[0]] = x 
    v[0] += 1
    return v

@njit
def v_pop(v):
    if(v[0] == 1): return 0
    v[0] -= 1
    return v[v[0]]


@njit
def add_pop_just_arr():
    v = v_new(10000)
    d = 0
    for i in range(10000):
        v_add(v,i)
    for i in range(10000):
        d = v_pop(v)

def add_pop_python_list():
    v = []
    for i in range(10000):
        v.append(i)

def add_pop_numpy_list():
    v = np.empty(10000, dtype=np.int64)
    for i in range(10000):
        v[i] = i

    # for i in range(10000):
    #     d = v_pop(v)
# test_vector()
# print(time_ms(add_pop))
# print(time_ms(add_pop_dd))
# print(time_ms(add_pop_manual))
# print(time_ms(add_pop_inline))
# print(time_ms(add_pop_just_arr))
# print(time_ms(add_pop_python_list))
# print(time_ms(add_pop_numpy_list))




# test_vector()







from cre.fact import define_fact
from cre.context import cre_context
from cre.transform.incr_processor import IncrProcessor, IncrProcessorType
from cre.memset import MemSet





def test_get_changes():
    with cre_context("test_get_changes"):
        BOOP = define_fact("BOOP",{"A": "string", "B" : "number"})

        ms = MemSet()
        ip = IncrProcessor(ms,0)

        b1 = BOOP("A", 1)
        b2 = BOOP("A", 1)
        b3 = BOOP("A", 1)
        ms.declare(b1)
        ms.declare(b2)
        ms.declare(b3)
        ms.modify(b3, "A", "Q")

        changes = ip.get_changes()
        assert len(changes) == 3
        assert changes[0].was_declared == True
        assert changes[1].was_declared == True
        assert changes[2].was_declared == True

        # Declarations clobber modifications
        assert changes[2].was_modified == False
        
        changes = ip.get_changes()
        # Get changes exhausts changes by default
        assert len(changes) == 0

        ms.modify(b1,"A", "Q")
        ms.modify(b1,"A", "V")
        ms.modify(b1,"B", 7)
        ms.modify(b2,"A", "Q")
        ms.modify(b3,"B", 7)
        
        changes = ip.get_changes()
        assert len(changes) == 3
        assert changes[0].was_modified == True
        assert len(changes[0].a_ids) == 2
        assert changes[1].was_modified == True
        assert len(changes[1].a_ids) == 1
        assert changes[2].was_modified == True
        assert len(changes[2].a_ids) == 1

        ms.retract(b1)
        ms.retract(b2)
        ms.retract(b3)

        b1 = BOOP("A", 1)
        b2 = BOOP("A", 1)
        b3 = BOOP("A", 1)

        ms.declare(b3)
        ms.declare(b2)
        ms.declare(b1)

        ms.modify(b1,"A", "Q")
        ms.modify(b1,"A", "V")
        ms.modify(b1,"B", 7)
        ms.modify(b2,"A", "Q")
        ms.modify(b3,"B", 7)

        changes = ip.get_changes()
        assert len(changes) == 3
        # If retraction and declaration both occur they are both true
        assert changes[0].was_declared == True
        assert changes[0].was_retracted == True
        assert changes[0].was_modified == False
        assert changes[1].was_declared == True
        assert changes[1].was_retracted == True
        assert changes[1].was_modified == False
        assert changes[2].was_declared == True
        assert changes[2].was_retracted == True
        assert changes[2].was_modified == False



if __name__ == "__main__":
    test_get_changes()



import numpy as np
from numba import f8, u1, u2, i8, u8, types, njit, generated_jit
from numba.types import FunctionType, unicode_type, ListType, boolean
from numba.typed import List
from numba.extending import  overload, overload_method
from cre.utils import cast, _load_ptr, _raw_ptr_from_struct, _raw_ptr_from_struct_incref, CastFriendlyMixin, decode_idrec, _func_from_address, _incref_structref, _struct_get_data_ptr, _sizeof_type, _decref_structref
from cre.structref import define_structref
from cre.obj import CREObjTypeClass, CREObjType, member_info_type, _iter_mbr_infos, OBJECT_MBR_ID, PRIMITIVE_MBR_ID
from numba.core.datamodel import default_manager, models
from numba.experimental.structref import define_attributes, StructRefProxy, new, define_boxing
import operator
from cre.core import T_ID_CONDITIONS, T_ID_LITERAL, T_ID_FUNC, T_ID_FACT, T_ID_VAR, T_ID_UNDEFINED, T_ID_BOOL, T_ID_INT, T_ID_FLOAT, T_ID_STR, T_ID_TUPLE_FACT
# from cre.primitive import BooleanPrimitiveType, IntegerPrimitiveType, FloatPrimitiveType, StringPrimitiveType
from cre.tuple_fact import TupleFact
from cre.var import VarType
# from cre.op import GenericOpType
from cre.func import CREFuncType, ARGINFO_VAR, ARGINFO_FUNC, ARGINFO_CONST
from cre.fact import BaseFact
from cre.conditions import LiteralType, ConditionsType
from numba.cpython.hashing import _Py_hash_t, _Py_uhash_t, _PyHASH_XXROTATE, _PyHASH_XXPRIME_1, _PyHASH_XXPRIME_2, _PyHASH_XXPRIME_5, process_return
from cre.hashing import accum_item_hash, unicode_hash_noseed



    



### __eq___ ###


@njit(boolean(u2,i8,i8), cache=True)
def eq_from_t_id_ptr(t_id, data_ptr_a, data_ptr_b):
    if(t_id == T_ID_BOOL):
        return _load_ptr(boolean, data_ptr_a) == _load_ptr(boolean, data_ptr_b)
    elif(t_id == T_ID_INT):
        return _load_ptr(i8, data_ptr_a) == _load_ptr(i8, data_ptr_b)
    elif(t_id == T_ID_FLOAT):
        return _load_ptr(f8, data_ptr_a) == _load_ptr(f8, data_ptr_b)
    elif(t_id == T_ID_STR):
        return _load_ptr(unicode_type, data_ptr_a) == _load_ptr(unicode_type, data_ptr_b)
    return False





@njit(boolean(CREObjType, CREObjType),cache=True)
def var_eq(a, b):
    t_id_b,_, _ = decode_idrec(b.idrec)
    if(t_id_b != T_ID_VAR): return False

    va = cast(a, VarType)
    vb = cast(b, VarType)

    if(va.base_ptr != vb.base_ptr): return False

    if(len(va.deref_infos) != len(vb.deref_infos)): return False
    for deref_info_a, deref_info_b in zip(va.deref_infos, vb.deref_infos):
        if(deref_info_a.offset != deref_info_b.offset): return False

    
    return True

@njit(boolean(CREObjType, CREObjType),cache=True)
def cre_func_eq(a, b):
    t_id_b,_, _ = decode_idrec(b.idrec)
    if(t_id_b != T_ID_FUNC): return False

    oa = cast(a, CREFuncType)
    ob = cast(b, CREFuncType)

    stack = List()
    stack.append((oa,ob))

    while(len(stack) > 0):
        oa,ob = stack.pop()
        if(oa.call_heads_addr != ob.call_heads_addr): return False
        if(oa.n_args != ob.n_args): return False
        if(len(oa.head_infos) != len(ob.head_infos)): return False

        infs_a = oa.root_arg_infos
        infs_b = ob.root_arg_infos

        for inf_a, inf_b in zip(infs_a, infs_b):

            if(inf_a.type != inf_b.type): return False
            if(inf_a.t_id != inf_b.t_id): return False
            if(inf_a.type == ARGINFO_VAR):
                va = cast(inf_a.ptr, VarType)
                vb = cast(inf_b.ptr, VarType)
                if(not var_eq(va,vb)): return False
            elif(inf_a.type == ARGINFO_CONST):
                is_eq = eq_from_t_id_ptr(inf_a.t_id, 
                    inf_a.ptr, inf_b.ptr)
                if(not is_eq): return False
            elif(inf_a.type == ARGINFO_FUNC):
                _oa = cast(inf_a.ptr, CREFuncType)
                _ob = cast(inf_b.ptr, CREFuncType)
                stack.append((_oa,_ob))

    return True

@njit(boolean(CREObjType, CREObjType),cache=True)
def literal_eq(a, b):
    t_id_b,_, _ = decode_idrec(b.idrec)
    if(t_id_b != T_ID_LITERAL): return False

    la = cast(a, LiteralType)
    lb = cast(b, LiteralType)

    if(not cre_func_eq(la.op, lb.op)): return False
    if(la.negated != lb.negated): return False
    
    return True

@njit(boolean(CREObjType, CREObjType),cache=True)
def conds_eq(a,b):
    t_id_b,_, _ = decode_idrec(b.idrec)
    if(t_id_b != T_ID_CONDITIONS): return False
    xa = cast(a, ConditionsType)
    xb = cast(b, ConditionsType)

    if(len(xa.dnf) != len(xb.dnf)): return False

    for conjuct_a, conjuct_b in zip(xa.dnf, xb.dnf):
        if(len(conjuct_a) != len(conjuct_b)): return False
        for lit_a, lit_b in zip(conjuct_a, conjuct_b):
            if(not literal_eq(lit_a, lit_b)): return False
    return True

@njit(boolean(CREObjType, CREObjType),cache=True)
def fact_eq(a, b):
    ''' based roughly on _tuple_hash from numba.cpython.hashing'''
    # while(not is_done):
    tla = a.num_chr_mbrs
    tlb = b.num_chr_mbrs

    if(cast(a, i8) == cast(b, i8)):
        return True

    if(tla != tlb): return False

    for (info_a), (info_b) in zip(_iter_mbr_infos(a),_iter_mbr_infos(b)):
        t_id_a, m_id_a, data_ptr_a = info_a
        t_id_b, m_id_b, data_ptr_b = info_b
        if(t_id_a == T_ID_UNDEFINED): 
            t_id_a,_, _ = decode_idrec(cast(_load_ptr(i8, data_ptr_a), CREObjType).idrec)
        if(t_id_b == T_ID_UNDEFINED): 
            t_id_b,_, _ = decode_idrec(cast(_load_ptr(i8, data_ptr_b), CREObjType).idrec)

        if(t_id_a != t_id_b): return False

        if(t_id_a == T_ID_TUPLE_FACT):
            raise Exception()
        elif(t_id_a == T_ID_FACT):
            if(not _load_ptr(i8, data_ptr_a) == _load_ptr(i8, data_ptr_b)): return False
        else:
            if(not eq_from_t_id_ptr(t_id_a, data_ptr_a, data_ptr_b)): return False

    
    return True


@njit(boolean(CREObjType, CREObjType),cache=True)
def tuple_fact_eq(a, b):
    ''' based roughly on _tuple_hash from numba.cpython.hashing'''
    t_id_b,_, _ = decode_idrec(b.idrec)
    if(t_id_b != T_ID_TUPLE_FACT): return False

    stack_buffer = None
    tf_a = cast(a, TupleFact)
    tf_b = cast(b, TupleFact)
    stack_head = -1
    is_done = False

    # acc = _PyHASH_XXPRIME_5
    while(not is_done):
        tla = tf_a.num_chr_mbrs
        tlb = tf_b.num_chr_mbrs

        if(tla != tlb): return False

        for (info_a), (info_b) in zip(_iter_mbr_infos(tf_a),_iter_mbr_infos(tf_b)):
            t_id_a, m_id_a, data_ptr_a = info_a
            t_id_b, m_id_b, data_ptr_b = info_b

            if(t_id_a == T_ID_UNDEFINED): 
                t_id_a,_, _ = decode_idrec(cast(_load_ptr(i8, data_ptr_a), CREObjType).idrec)
            if(t_id_b == T_ID_UNDEFINED): 
                t_id_b,_, _ = decode_idrec(cast(_load_ptr(i8, data_ptr_b), CREObjType).idrec)


            if(t_id_a != t_id_b): return False

            if(t_id_a == T_ID_TUPLE_FACT):
                # Use an inline stack instead of recrusion because numba breaks on recursion
                stack_head += 1
                if(stack_buffer is None):
                    stack_buffer = np.empty((2,2), dtype=np.int64)
                if(stack_head >= len(stack_buffer)):
                    new_stack_buffer  = np.empty(((stack_head+1)*2,2), dtype=np.int64)
                    new_stack_buffer[:len(stack_buffer)] = stack_buffer
                    stack_buffer = new_stack_buffer
                stack_buffer[stack_head,0] = _load_ptr(i8, data_ptr_a)
                stack_buffer[stack_head,1] = _load_ptr(i8, data_ptr_b)

            elif(t_id_a <= T_ID_STR):
                if(not eq_from_t_id_ptr(t_id_a, data_ptr_a, data_ptr_b)):
                    return False
            else:
                # Kind of a heavy handed way to do this... but lack
                #  of recrusion makes it the best option                    
                mbr_a = cast(_load_ptr(i8,data_ptr_a), CREObjType)
                mbr_b = cast(_load_ptr(i8,data_ptr_b), CREObjType)
                if(t_id_a==T_ID_VAR):
                    if(not var_eq(mbr_a,mbr_b)): return False
                elif(t_id_a==T_ID_FUNC):
                    if(not cre_func_eq(mbr_a,mbr_b)): return False
                elif(t_id_a==T_ID_LITERAL):
                    if(not literal_eq(mbr_a,mbr_b)): return False
                elif(t_id_a==T_ID_CONDITIONS):
                    if(not conds_eq(mbr_a,mbr_b)): return False
                else:
                    if(not fact_eq(mbr_a,mbr_b)): return False

                

        if(stack_head > -1):
            tf_a = cast(stack_buffer[stack_head,0], TupleFact);
            tf_b = cast(stack_buffer[stack_head,1], TupleFact);
            stack_head -=1;
        else:
            is_done = True

    return True



@overload(operator.eq)
def _cre_obj_eq(a,b):
    # print(a,b, isinstance(a,CREObjTypeClass) and isinstance(b,CREObjTypeClass))
    if(isinstance(a,CREObjTypeClass) and isinstance(b,CREObjTypeClass)):
        def impl(a, b):
            t_id,_, _ = decode_idrec(a.idrec)
            if(t_id==T_ID_TUPLE_FACT):
                return tuple_fact_eq(a,b)
            elif(t_id==T_ID_VAR):
                return var_eq(a,b)
            elif(t_id==T_ID_FUNC):
                return cre_func_eq(a,b)
            elif(t_id==T_ID_LITERAL):
                return literal_eq(a,b)
            elif(t_id==T_ID_CONDITIONS):
                return conds_eq(a,b)
            else:
                return fact_eq(a,b)
            
        return impl

@njit(boolean(CREObjType,CREObjType), cache=True)
def cre_obj_eq(a,b):
    return a == b


### __hash___ ###


# print(_PyHASH_XXPRIME_1, _PyHASH_XXPRIME_2, _PyHASH_XXPRIME_5)

@njit(_Py_uhash_t(u2,i8), cache=True)
def hash_from_t_id_ptr(t_id, data_ptr):
    if(t_id == T_ID_BOOL):
        return hash(_load_ptr(boolean, data_ptr))
    elif(t_id == T_ID_INT):
        return hash(_load_ptr(i8, data_ptr))
    elif(t_id == T_ID_FLOAT):
        return hash(_load_ptr(f8, data_ptr))
    elif(t_id == T_ID_STR):
        return unicode_hash_noseed(_load_ptr(unicode_type, data_ptr))
    # print(t_id, "BAD T_ID")
    return u8(0)













@njit(_Py_hash_t(CREObjType),cache=True)
def var_hash(x):
    ''' based roughly on _tuple_hash from numba.cpython.hashing'''
    # while(not is_done):
    if(x.hash_val == 0):
        vx = cast(x, VarType)

        acc = _PyHASH_XXPRIME_5
        acc = accum_item_hash(acc,  vx.base_ptr) 
        acc = accum_item_hash(acc, len(vx.deref_infos))
        
        for deref_info in vx.deref_infos:
            acc = accum_item_hash(acc, deref_info.offset) 
        x.hash_val = _Py_hash_t(acc)

    
    return x.hash_val


@njit(_Py_hash_t(CREObjType),cache=True)
def cre_func_hash(x):
    if(x.hash_val == 0):
        ox = cast(x, CREFuncType)
        stack = List()
        stack.append(ox)

        acc = _PyHASH_XXPRIME_5

        while(len(stack) > 0):
            ox = stack.pop()
            acc = accum_item_hash(acc, ox.call_heads_addr) 
            acc = accum_item_hash(acc, ox.n_args)
            
            for inf in ox.root_arg_infos:
                if(inf.type == ARGINFO_VAR):
                    _var = cast(inf.ptr, VarType)
                    acc = accum_item_hash(acc, var_hash(_var)) 
                elif(inf.type == ARGINFO_CONST):
                    const_hash = hash_from_t_id_ptr(inf.t_id, inf.ptr)
                    acc = accum_item_hash(acc, const_hash) 
                elif(inf.type == ARGINFO_FUNC):
                    _ox = cast(inf.ptr, CREFuncType)
                    stack.append(_ox)
        x.hash_val = acc
    return x.hash_val



@njit(_Py_hash_t(CREObjType),cache=True)
def literal_hash(x):
    if(x.hash_val == 0):
        lx = cast(x, LiteralType)

        acc = cre_func_hash(lx.op)
        acc = accum_item_hash(acc, lx.negated) 

        x.hash_val = acc

    return x.hash_val    


@njit(_Py_hash_t(CREObjType),cache=True)
def conds_hash(x):
    if(x.hash_val == 0):
        cx = cast(x, ConditionsType)

        acc = _PyHASH_XXPRIME_5
        for conjuct in cx.dnf:
            acc = accum_item_hash(acc, _PyHASH_XXPRIME_2)            
            for lit in conjuct:
                acc = accum_item_hash(acc, literal_hash(lit))
        x.hash_val = acc

    return x.hash_val    

@njit(_Py_hash_t(CREObjType),cache=True)
def fact_hash(x):
    ''' based roughly on _tuple_hash from numba.cpython.hashing'''
    if(x.hash_val == 0):
        acc = _PyHASH_XXPRIME_5
        tl = x.num_chr_mbrs
        for t_id, m_id, data_ptr in _iter_mbr_infos(x):
            if(t_id == T_ID_UNDEFINED): 
                t_id,_, _ = decode_idrec(cast(_load_ptr(i8, data_ptr), CREObjType).idrec)
            if(t_id == T_ID_TUPLE_FACT):
                raise Exception()
                # acc = accum_item_hash(acc, tuple_fact_hash(x))
            elif(m_id == PRIMITIVE_MBR_ID):
                # hash primitives
                acc = accum_item_hash(acc, hash_from_t_id_ptr(t_id, data_ptr))                
            else:
                # hash on the pointer 
                # ptr = _load_ptr(i8, data_ptr)
                # acc = accum_item_hash(acc, _PyHASH_XXPRIME_5 if(ptr == 0) else ptr)
                pass

        acc = accum_item_hash(acc,tl)
        x.hash_val = acc


    return x.hash_val

@njit(_Py_hash_t(CREObjType),cache=True)
def tuple_fact_hash(x):
    ''' based roughly on _tuple_hash from numba.cpython.hashing'''
    if(x.hash_val == 0):
        stack_buffer = None
        p = cast(x, TupleFact)
        stack_head = -1
        is_done = False

        acc = _PyHASH_XXPRIME_5
        while(not is_done):
            tl = p.num_chr_mbrs
            for t_id, m_id, data_ptr in _iter_mbr_infos(p):
                # print(":: t_id", t_id, data_ptr)
                if(t_id == T_ID_UNDEFINED): 
                    t_id,_, _ = decode_idrec(cast(_load_ptr(i8, data_ptr), CREObjType).idrec)
                if(t_id == T_ID_TUPLE_FACT):
                    # print("IS PRED")
                    # Use an inline stack instead of recrusion because numba can't cache recursion
                    stack_head += 1
                    if(stack_buffer is None):
                        stack_buffer = np.empty(2, dtype=np.int64)
                    if(stack_head >= len(stack_buffer)):
                        new_stack_buffer  = np.empty((stack_head+1)*2, dtype=np.int64)
                        new_stack_buffer[:len(stack_buffer)] = stack_buffer
                        stack_buffer = new_stack_buffer
                    stack_buffer[stack_head] = _load_ptr(i8, data_ptr)

                elif(t_id <= T_ID_STR):
                    acc = accum_item_hash(acc, hash_from_t_id_ptr(t_id, data_ptr))
                    
                else:
                    # Kind of a heavy handed way to do this... but lack
                    #  of recrusion makes it the best option                    
                    mbr = cast(_load_ptr(i8,data_ptr), CREObjType)
                    # print("::  " ,decode_idrec(x.idrec), x)
                    if(t_id==T_ID_VAR):
                        # print("----- VAR -----")
                        mbr_hash = var_hash(mbr)
                    elif(t_id==T_ID_FUNC):
                        # print("----- OP -----")
                        mbr_hash = cre_func_hash(mbr)
                    elif(t_id==T_ID_LITERAL):
                        mbr_hash = literal_hash(mbr)
                    elif(t_id==T_ID_CONDITIONS):
                        mbr_hash = conds_hash(mbr)
                    else:
                        mbr_hash = fact_hash(mbr)
                    acc = accum_item_hash(acc,  mbr_hash )
                    # print(t_id, "<<", acc,hash_from_t_id_ptr(t_id, data_ptr))

            acc = accum_item_hash(acc,tl)

            if(stack_head > -1):
                p = cast(stack_buffer[stack_head], TupleFact);
                stack_head -=1;
            else:
                is_done = True
        x.hash_val = acc

    return x.hash_val

# @njit(_Py_hash_t(CREObjType))
# def cre_obj_hash(x):
#     t_id,_,_ = decode_idrec(x.idrec)

#     # if(t_id <= T_ID_STR):
#     #     hsh = non_tuple_fact_hash(x)
#     # elif(t_id == T_ID_TUPLE_FACT):
#     hsh = tuple_fact_hash(x)

#     # return i8()
#     if hsh == _Py_uhash_t(-1):
#         return process_return(1546275796)

#     return process_return(hsh)

@njit(_Py_hash_t(CREObjType), locals={"hsh" : _Py_hash_t}, cache=True)
def cre_obj_hash(x):
    t_id,_, _ = decode_idrec(x.idrec)
    hsh = 0
    if(t_id==T_ID_TUPLE_FACT):
        hsh = tuple_fact_hash(x) 
    elif(t_id==T_ID_VAR):
        hsh = var_hash(x) 
    elif(t_id==T_ID_FUNC):
        hsh = cre_func_hash(x) 
    elif(t_id==T_ID_LITERAL):
        hsh = literal_hash(x)
    elif(t_id==T_ID_CONDITIONS):
        hsh = conds_hash(x)
    else:
        hsh = fact_hash(x)
    return hsh



@overload(hash)
@overload_method(CREObjTypeClass, '__hash__')
@overload_method(CREObjType, '__hash__') # Not sure why but is necessary
def _cre_obj_hash_overload(x):
    print("------IMPL-----", x, isinstance(x, CREObjTypeClass))
    if(isinstance(x,CREObjTypeClass)):
        def impl(x):
            return cre_obj_hash(x)
        return impl


#### __str__ ### 


# def __cre_obj_hash(x):
#     if(isinstance(x,CREObjTypeClass)):
#         def impl(x):
#             return cre_obj_hash(x)
#         return impl





import numpy as np
from numba import njit, generated_jit, types, literal_unroll, u8, i8, f8, u1, u2
from numba.types import unicode_type,  intp, Tuple,  Tuple, DictType, ListType
from numba.typed import Dict, List
from numba.experimental.structref import new
from cre.core import short_name
from cre.obj import CREObjType
from cre.fact import define_fact, UntypedFact, call_untyped_fact, BaseFact
from cre.fact_intrinsics import fact_lower_getattr, resolve_fact_getattr_type
from cre.tuple_fact import TupleFact, TF
from cre.context import cre_context, CREContext
# from cre.default_ops import Add, Subtract, Divide
from cre.var import Var, VarType, var_append_deref
# from cre.op import GenericOpType
from cre.utils import _dict_from_ptr, ptr_t, _func_from_address, _obj_cast_codegen, _func_from_address, _incref_structref, _ptr_from_struct_incref
from cre.structref import define_structref
from cre.transform.incr_processor import incr_processor_fields, IncrProcessorType, init_incr_processor
from cre.structref import CastFriendlyStructref, define_boxing
from numba.experimental import structref
from numba.extending import overload_method, overload, lower_cast, SentryLiteralArgs
from numba.experimental.function_type import _get_wrapper_address
import cloudpickle
from cre.gval import get_gval_type, new_gval, gval as gval_type
from numba.core.typing.typeof import typeof


ptr_t_pair = Tuple((ptr_t,ptr_t))
enumerizer_fields = {
    # For each t_id a pair of dictionaries: 1) maps values to u8 enums 2) maps u8 enums to values.
    "val_maps" : DictType(u2, ptr_t_pair),
}

@structref.register
class EnumerizerTypeClass(CastFriendlyStructref):
    pass

EnumerizerType = EnumerizerTypeClass([(k,v) for k,v in enumerizer_fields.items()])

@njit(cache=True)    
def enumerizer_ctor():
    st = new(EnumerizerType)
    st.val_maps = Dict.empty(u2, ptr_t_pair)
    return st
    
class Enumerizer(structref.StructRefProxy):
    def __new__(cls):
        self = enumerizer_ctor()
        return self

    def dict_for_type(self,val_type):
        return dicts_for_type(self,val_type)

    def to_enum(self, val, d=None, inv_d=None):
        return to_enum(self, val, d, inv_d)

    def from_enum(self, enum, val_type, d=None, inv_d=None):
        return from_enum(self, enum, val_type, d, inv_d)


# # Classes that mimic behavior of a dictionary 

# class EnumerizerMap:
#     def __init__(self, enumerizer):
#         self.enumerizer = enumerizer

#     def __getitem__(self, val):
#         self.enumerizer.to_enum(val)

# class EnumerizerInvMapper(EnumerizerMap):
#     def __init__(self, enumerizer):
#         self.enumerizer = enumerizer

#     def __call__(self, enum, val_type):
        

#     def __getitem__(self, enum, val_type):
#         self.enumerizer.from_enum(enum, val_type)



define_boxing(EnumerizerTypeClass, Enumerizer)

@generated_jit(cache=True,nopython=True)
@overload_method(EnumerizerTypeClass, 'dicts_for_type')
def dicts_for_type(self, val_type):
    val_type = val_type.instance_type
    context = CREContext.get_default_context()
    t_id = u2(context.get_t_id(_type=val_type))
    d_type = DictType(val_type,u8)
    inv_d_type = DictType(u8, val_type)
    def impl(self, val_type):
        if(t_id not in self.val_maps):
            d = Dict.empty(val_type,u8)
            inv_d = Dict.empty(u8, val_type)
            d_ptr, inv_d_ptr = _ptr_from_struct_incref(d), _ptr_from_struct_incref(inv_d)
            self.val_maps[t_id] = (d_ptr, inv_d_ptr)

        d_ptr, inv_d_ptr = self.val_maps[t_id]
        d = _dict_from_ptr(d_type, d_ptr)
        inv_d = _dict_from_ptr(inv_d_type, inv_d_ptr)
        return d, inv_d
    return impl

# @generated_jit(cache=True,nopython=True)
# @overload_method(EnumerizerTypeClass, 'inv_dict_for_type')
# def inv_dict_for_type(self, val_type):
#     val_type = val_type.instance_type
#     context = CREContext.get_default_context()
#     t_id = u2(context.get_t_id(_type=val_type))
#     d_type = DictType(u8, val_type)
#     def impl(self, val_type):
#         if(t_id not in self.inv_val_maps):
#             d = Dict.empty(u8, val_type)
#             self.inv_val_maps[t_id] = _ptr_from_struct_incref(d)
#         d = _dict_from_ptr(d_type, self.inv_val_maps[t_id])
#         return d
#     return impl

@generated_jit(cache=True,nopython=True)
@overload_method(EnumerizerTypeClass, 'to_enum')
def to_enum(self, val, d=None, inv_d=None):
    val_type = val
    def impl(self, val, d=None, inv_d=None):
        if(d is None or inv_d is None):
            d,inv_d = dicts_for_type(self, val_type)
        if(val not in d):
            enum = u8(len(d)+1)
            d[val] = enum
            inv_d[enum] = val
        return d[val]
    return impl

@generated_jit(cache=True,nopython=True)
@overload_method(EnumerizerTypeClass, 'from_enum')
def from_enum(self, enum, val_type, d=None, inv_d=None):
    # val_type = val_type.instance_type
    def impl(self, enum, val_type, d=None, inv_d=None):
        if(d is None or inv_d is None):
            d, inv_d = dicts_for_type(self, val_type)

        if(enum not in inv_d):
            raise ValueError("No value found for enum.")

        val = inv_d[enum]
        return val
    return impl

# @



# e = Enumerizer()
# print(e.enumerize(6.))
# print(e.enumerize(8.))
# print(e.enumerize(1.))
# print(e.enumerize(1.))
# print(e.enumerize(True))
# print(e.enumerize("A"))




from cre.fact import define_fact
from cre.structref import define_structref
from numba import njit,literally
from numba.types import unicode_type, f8, i8, ListType
from numba.typed import List
from numba.experimental.structref import new
import logging
import numpy as np
import pytest

from cre.utils import (_incref_structref, _decref_structref, 
         _meminfo_from_struct, _struct_from_meminfo,  _raw_ptr_from_struct,
         _struct_from_ptr, encode_idrec, decode_idrec, _struct_get_attr_offset, 
         _struct_get_data_ptr, _load_ptr, struct_get_attr_offset, 
         _listtype_sizeof_item, _ptr_from_struct_incref, _list_base_from_ptr,
         cast)
from numba.core.runtime.nrt import rtsys

BOOP, BOOPType = define_structref("BOOP", [("A", unicode_type), ("B", i8)])


##### test_encode_decode #####

def test_encode_decode():
    idrec = encode_idrec(7,8,9)
    assert isinstance(idrec,int)
    assert decode_idrec(idrec) == (7,8,9)

    idrec = encode_idrec(0x0000,1099511627775,255)
    assert isinstance(idrec,int)
    assert decode_idrec(idrec) == (0x0000,1099511627775,255)

    idrec = encode_idrec(0xFFFF,1099511627775,255)
    assert isinstance(idrec,int)
    assert decode_idrec(idrec) == (0xFFFF,1099511627775,255)


#### test_structref_to_meminfo ####

@njit(cache=True)
def new_boop():
    return BOOP("?", 1)

@njit(cache=True)
def rip_meminfo_inc(st):
    # _incref_structref(st)
    return _meminfo_from_struct(st)

@njit(cache=True)
def restore_struct_from_meminfo(m):
    return _struct_from_meminfo(BOOPType,m)

def test_structref_to_meminfo():
    b1 = new_boop()
    b2 = new_boop.py_func()

    assert b1._meminfo.refcount == 1
    assert b2._meminfo.refcount == 1

    meminfo1 = rip_meminfo_inc(b1)
    meminfo2 = rip_meminfo_inc(b2)

    assert b1._meminfo.refcount == 2
    assert b2._meminfo.refcount == 2

    b1 = restore_struct_from_meminfo(meminfo1)
    b2 = restore_struct_from_meminfo(meminfo2)

    meminfo1, meminfo2 = None, None

    assert b1.A == "?" and b1.B == 1
    assert b2.A == "?" and b2.B == 1


#### test_structref_to_pointer ####

@njit(cache=True)
def clear_incref(st):
    _decref_structref(st)

@njit(cache=True)
def rip_pointer_inc(st):
    _incref_structref(st)
    return _raw_ptr_from_struct(st)

@njit(cache=True)
def restore_struct_from_ptr(p):
    return _struct_from_ptr(BOOPType,p)


def test_structref_to_pointer():
    b1 = new_boop()
    b2 = new_boop.py_func()

    assert b1._meminfo.refcount == 1
    assert b2._meminfo.refcount == 1

    ptr1 = rip_pointer_inc(b1)
    ptr2 = rip_pointer_inc(b2)

    assert b1._meminfo.refcount == 2
    assert b2._meminfo.refcount == 2

    b1 = restore_struct_from_ptr(ptr1)
    b2 = restore_struct_from_ptr(ptr2)

    assert b1.A == "?" and b1.B == 1
    assert b2.A == "?" and b2.B == 1

    clear_incref(b1)
    clear_incref(b2)

    assert b1._meminfo.refcount == 1
    assert b2._meminfo.refcount == 1

# @njit(cache=True)
# def struct_get_attr_offset(b,attr):
#     return _struct_get_attr_offset(b,literally(attr))

@njit(cache=True)
def load_offset(typ, b ,offset):
    return _load_ptr(typ,_struct_get_data_ptr(b)+offset)



BEEP1, BEEP1Type = define_structref("BEEP1", [("A", i8), ("B", i8),("C", i8)])
BEEP2, BEEP2Type = define_structref("BEEP2", [("A", unicode_type), ("B", i8),("C", i8)])

def test_direct_member_access():
    b1 = BEEP1(1,2,1)
    b2 = BEEP1(1,2,3)

    assert struct_get_attr_offset(b1,"A") == 0
    assert struct_get_attr_offset(b1,"B") == 8
    assert struct_get_attr_offset(b1,"C") == 16

    offset = struct_get_attr_offset(b1,"C")
    cls_offset = struct_get_attr_offset(BEEP1Type,"C")

    assert offset == cls_offset

    assert load_offset(i8,b2,offset) == 3

    b1 = BEEP2("b1",2,1)
    b2 = BEEP2("b2",2,3)

    offset = struct_get_attr_offset(b1,"C")
    cls_offset = struct_get_attr_offset(BEEP2Type,"C")
    assert offset == cls_offset

    offset = struct_get_attr_offset(b1,"A")
    cls_offset = struct_get_attr_offset(BEEP2Type,"A")
    assert offset == cls_offset

    load_offset(unicode_type,b2,offset) == "b2"


@njit(cache=True)
def listtype_sizeof(lst_typ):
    return _listtype_sizeof_item(lst_typ)

@njit(cache=True)
def get_3(ptr,typ,lst_typ):
    item_size = _listtype_sizeof_item(lst_typ)
    base_ptr = _list_base_from_ptr(ptr)
    a = _load_ptr(typ,base_ptr)
    b = _load_ptr(typ,base_ptr+item_size*1)
    c = _load_ptr(typ,base_ptr+item_size*2)
    return a,b,c

unc_lst = ListType(unicode_type)
@njit(cache=True)
def _test_list_getitem_intrinsic_unicode_type():
    l = List.empty_list(unicode_type)
    l.append("A")
    l.append("B")
    l.append("C")
    ptr = _ptr_from_struct_incref(l)
    a,b,c = get_3(ptr,unicode_type,unc_lst)
    return a, b, c

f8_lst = ListType(f8)
@njit(cache=True)
def _test_list_getitem_intrinsic_f8():
    l = List.empty_list(f8)
    l.append(1)
    l.append(2)
    l.append(3)
    ptr = _ptr_from_struct_incref(l)
    a,b,c = get_3(ptr,f8,f8_lst)
    return a, b, c

BOOP_lst = ListType(BOOPType)
@njit(cache=True)
def _test_list_getitem_intrinsic_BOOP():
    l = List.empty_list(BOOPType)
    l.append(BOOP("A",1))
    l.append(BOOP("B",2))
    l.append(BOOP("C",3))
    ptr = _ptr_from_struct_incref(l)
    a,b,c = get_3(ptr,BOOPType,BOOP_lst)
    return a, b, c
    
def test_list_intrinsics():
    assert listtype_sizeof(unc_lst) == 48 
    assert listtype_sizeof(f8_lst) == 8 
    assert listtype_sizeof(BOOP_lst) == 8 #should be ptr width

    assert _test_list_getitem_intrinsic_unicode_type() == ("A","B","C")
    assert _test_list_getitem_intrinsic_f8() == (1,2,3)
    assert [x.A for x in _test_list_getitem_intrinsic_BOOP()] == ["A","B","C"]


@njit(cache=True)
def _test_cast_obj():
    b = BOOP("A",1)
    ptr = cast(b, i8)
    c = cast(ptr, BOOPType)
    d = cast(b, BOOPType)
    return b.A + d.A


@njit(cache=True)
def _test_cast_list():
    l = List.empty_list(unicode_type)
    l.append("A")
    l.append("B")
    l.append("C")
    ptr = cast(l, i8)
    m = cast(ptr, unc_lst)
    return m[1]

def test_cast():
    assert _test_cast_obj() == "AA"
    assert _test_cast_list() == "B"


###################### BENCHMARKS ########################

#### b_encode_idrec ####

def gen_rand_nums():
    return (np.random.randint(1000,size=(10000,3)),), {}

@njit(cache=True)
def _b_encode_idrec(rand_nums):
    for x in rand_nums:
        encode_idrec(x[0],x[1],x[2])

@pytest.mark.benchmark(group="utils")
def test_b_encode_idrec(benchmark):
    benchmark.pedantic(_b_encode_idrec,setup=gen_rand_nums, warmup_rounds=1)


#### b_decode_idrec ####

def gen_rand_idrecs():
    return (np.random.randint(0xFFFFFFFF,size=(10000,),dtype=np.uint64),), {}

@njit(cache=True)
def _b_decode_idrec(rand_idrecs):
    for x in rand_idrecs:
        decode_idrec(x)

@pytest.mark.benchmark(group="utils")
def test_b_decode_idrec(benchmark):
    benchmark.pedantic(_b_decode_idrec,setup=gen_rand_idrecs, warmup_rounds=1)



# print(_test_list_intrinsics_unicode())
# print(_test_list_intrinsics_f8())






    # _load_ptr(i8,_struct_get_data_pointer(b2)+offset) == 3


    




if __name__ == "__main__":
    test_cast()
    # test_structref_to_meminfo()
    # test_structref_to_pointer()
    # test_direct_member_access()
    # test_list_intrinsics()









    



import numpy as np
from numba import njit, f8
from numba.types import unicode_type, boolean
from cre.utils import decode_idrec 
from cre.context import cre_context 
from cre.memset import MemSet 
from cre.transform.flattener import Flattener, flattener_update
from cre.fact import define_fact
import pytest_benchmark
from cre.default_funcs import Equals
from cre.transform.feature_applier import FeatureApplier
from cre.transform.vectorizer import Vectorizer
from pprint import pprint


eq_f8 = Equals(f8, f8)
eq_str = Equals(unicode_type, unicode_type)


def test_vectorizer():
    with cre_context("test_vectorizer") as context:
        spec1 = {"id" : {"type" : "string", "visible" : False},
                 "u" : {"type" : "string", "visible" : True}, 
                 "v" : {"type" : "number", "visible" : False}}
        BOOP1 = define_fact("BOOP1", spec1)
        spec2 = {"inherit_from" : BOOP1,
                 "q" : {"type" : "number", "visible" : True}}
        BOOP2 = define_fact("BOOP2", spec2)
        spec3 = {"inherit_from" : BOOP2,
                 "x" : {"type" : "number", "visible" : True}}
        BOOP3 = define_fact("BOOP3", spec3)
   
        fa = FeatureApplier([eq_f8,eq_str],MemSet())
        vr = Vectorizer((f8,unicode_type,boolean))

        ms = MemSet()
        a = BOOP1(id="A",u="1", v=1)
        b = BOOP1(id="B",u="2", v=2)
        c = BOOP2(id="C",u="3", v=3, q=13)
        d = BOOP2(id="D",u="1", v=4, q=13)
        e = BOOP3(id="E",u="2", v=5, q=14, x=106)
        f = BOOP3(id="F",u="3", v=6, q=16, x=106)

        a_idrec = ms.declare(a)
        b_idrec = ms.declare(b)
        c_idrec = ms.declare(c)
        d_idrec = ms.declare(d)
        e_idrec = ms.declare(e)
        f_idrec = ms.declare(f)
        print("-------")

        fl = Flattener((BOOP1, BOOP2, BOOP3), ms, id_attr="id")
        flat_ms = fl()

        fa = FeatureApplier([eq_f8, eq_str],flat_ms)
        feat_ms = fa()
        # print(feat_ms)
        
        facts = feat_ms.get_facts()
        orig_len = len(facts)
        # for fact in facts:
            # print(fact)
        print("LEN", len(facts))
        # print(feat_ms)    
        # print()

        
        # Make sure that the vector has same size as set of gval facts
        floats, noms = vr(feat_ms)

        print(noms)

        assert len(noms) == orig_len
        assert not np.any(noms == 0)

    
        # Make sure that there are holes when retract facts
        ms.retract(c)
        ms.retract(d)

        flat = fl()
        feat = fa()
        floats, noms = vr(feat_ms)

        assert len(noms) == orig_len
        assert np.any(noms == 0)

        print(noms)

        # Make sure that vectors grow when add new facts
        ms.declare(BOOP3(id="Z",u="q", v=77, q=16, x=106))
        ms.declare(BOOP3(id="W",u="q", v=77, q=17, x=108))

        flat = fl()
        feat = fa()
        floats, noms = vr(feat_ms)

        print(noms)

        assert len(noms) > orig_len
        assert np.any(noms == 0)

        # Make sure that if we refill holes with equivalent facts that 
        #  the vectors' holes are filled
        ms.declare(BOOP2(id="C",u="3", v=3, q=13))
        ms.declare(BOOP2(id="D",u="1", v=4, q=13))

        flat = fl()
        feat = fa()
        floats, noms = vr(feat_ms)

        print(noms)

        assert len(noms) > orig_len
        assert not np.any(noms == 0)




        # Make sure that get_inv_map works
        # inv_map = vr.get_inv_map()





        # pprint({k:v for k,v in inv_map.items()})


        # continuous, nominal = vr(feat_ms)
        # print(continuous, nominal)
        # return

        # ms.retract(c)
        # ms.retract(d)

        # flat_ms = fl()
        # feat_ms = fa()
        # print(flat_ms)
        # print(feat_ms)
        # print(len(feat_ms.get_facts()))
        # print(count_true_false(feat_ms))


        # assert len(feat_ms.get_facts()) == ((3*4) + (3*4))
        # assert count_true_false(feat_ms)[0] == 4

        # ms.modify(e, "x", 777)
        # ms.modify(f, "u", "Z")

        # flat_ms = fl()
        # feat_ms = fa()
        # print(flat_ms)
        # print(feat_ms)
        # print(len(feat_ms.get_facts()))

        # print()
        # assert len(feat_ms.get_facts()) == ((3*4) + (3*4))
        # assert count_true_false(feat_ms)[0] == 2

if(__name__ == "__main__"):
    import faulthandler; faulthandler.enable()
    test_vectorizer()



import operator
import numpy as np
import numba
from numba import types, njit, i8, u8, i4, u1,u2,u4,  i8, literally, generated_jit, objmode
from numba.typed import List
from numba.core.types import ListType, unicode_type, void, Tuple
from numba.experimental import structref
from numba.experimental.structref import new, define_attributes, _Utils
from numba.extending import SentryLiteralArgs, lower_cast, overload_method, intrinsic, overload_attribute, intrinsic, lower_getattr_generic, overload, infer_getattr, lower_setattr_generic
from numba.core.typing.templates import AttributeTemplate
from cre.caching import gen_import_str, unique_hash,import_from_cached, source_to_cache, source_in_cache
from cre.context import cre_context
from cre.structref import define_structref, define_boxing, define_structref_template, CastFriendlyStructref
from cre.fact import define_fact, BaseFact, cast_fact, DeferredFactRefType, Fact, _standardize_type
from cre.utils import PrintElapse, ptr_t, _struct_from_meminfo, _meminfo_from_struct, _cast_structref, cast_structref, decode_idrec, lower_getattr, _struct_from_ptr,  lower_setattr, lower_getattr, _raw_ptr_from_struct, _decref_ptr, _incref_ptr, _incref_structref, _ptr_from_struct_incref
from cre.utils import assign_to_alias_in_parent_frame, encode_idrec, _obj_cast_codegen
from cre.vector import VectorType
from cre.obj import cre_obj_field_dict,CREObjType, CREObjTypeClass, CREObjProxy, set_chr_mbrs
# from cre.predicate_node import BasePredicateNode,BasePredicateNodeType, get_alpha_predicate_node_definition, \
 # get_beta_predicate_node_definition, deref_attrs, define_alpha_predicate_node, define_beta_predicate_node, AlphaPredicateNode, BetaPredicateNode
from numba.core import imputils, cgutils
from numba.core.datamodel import default_manager, models


from operator import itemgetter
from copy import copy
from os import getenv
from cre.utils import deref_info_type, DEREF_TYPE_ATTR, DEREF_TYPE_LIST, listtype_sizeof_item, _obj_cast_codegen
from cre.core import T_ID_RULE, register_global_default
from cre.conditions import ConditionsType, Conditions
from cre.memset import MemSetType
import cloudpickle
import inspect

rhs_ptrs_type = types.FunctionType(types.void(MemSetType, i8[::1]))

rule_fields_dict = {
    **cre_obj_field_dict,
    'lhs' : ConditionsType,
    'decl_rhs' : types.optional(ConditionsType),
    'imper_rhs' : types.optional(rhs_ptrs_type),
}


class RuleTypeClass(CREObjTypeClass):
    t_id = T_ID_RULE

    def __str__(self):
        return "cre.RuleType"

default_manager.register(RuleTypeClass, models.StructRefModel)
define_attributes(RuleTypeClass)

RuleType = RuleTypeClass([(k,v) for k,v in rule_fields_dict.items()])
register_global_default("Rule", RuleType)


class RuleMeta(type):
    def __call__(cls, *args, **kwargs):
        lhs = args[0]
        if(len(args)==1):
            def wrapper(rhs_func):
                rule = new_rule(lhs, rhs_func)
                return rule
            return wrapper
        else:
            return new_rule(*args)


class Rule(CREObjProxy,metaclass=RuleMeta):
    t_id = T_ID_RULE
    def __new__(cls, lhs, rhs):
        return new_rule(lhs, rhs)
    def __str__(self):
        return f"Rule(?)"
    def __repr__(self):
        return f"Rule(?)"
        # print("Rule.__new__", lhs, rhs)


define_boxing(RuleTypeClass,Rule)


@njit(RuleType(ConditionsType),cache=True)
def rule_ctor(lhs):
    st = new(RuleType)
    st.lhs = lhs
    return st

def gen_rhs_source(match_types, rhs_pyfunc):
    sig = types.void(MemSetType, *match_types)

    source = \
f'''from numba import njit, i8 
from numba import types
from cre.utils import _struct_tuple_from_pointer_arr
from cre.memset import MemSetType
import cloudpickle
# from numba.experimental.function_type import _get_wrapper_address

match_types = cloudpickle.loads({cloudpickle.dumps(match_types)})
rhs_pyfunc = cloudpickle.loads({cloudpickle.dumps(rhs_pyfunc)})

print(match_types)
signature = types.void(MemSetType, *match_types)
print(signature)
rhs = njit(signature, cache=True)(rhs_pyfunc)

@njit(types.void(MemSetType, i8[::1]), cache=True)
def rhs_ptrs(wm, ptrs):
    args = _struct_tuple_from_pointer_arr(match_types, ptrs)
    rhs(wm, *args)

# rhs_ptrs_addr = _get_wrapper_address(rhs_ptrs, types.void(MemSetTypem, i8[::1]))

'''
    return source


def new_rule(lhs, rhs):
    with PrintElapse("rule_ctor"):
        st = rule_ctor(lhs)

    if(isinstance(rhs, Conditions)):
        print("!", rhs)

    elif(hasattr(rhs,'__call__')):
        with PrintElapse("make source"):
            match_types = tuple([x.base_type for x in lhs.vars])
            func_src = inspect.getsource(rhs)
            long_hash = unique_hash([match_types, func_src])
            if(not source_in_cache('Rule_RHS', long_hash)):
                source = gen_rhs_source(match_types, rhs)
                source_to_cache('Rule_RHS', long_hash, source)

        with PrintElapse("run"):
            rhs_ptrs = import_from_cached('Rule_RHS', long_hash, ['rhs_ptrs'])['rhs_ptrs']
            print(list(rhs_ptrs.overloads.keys()))
        with PrintElapse("assign"):
            rule_assign_imper_rhs(st, rhs_ptrs)
            st._rhs_pyfunc = rhs
        return st


@njit(cache=True)
def rule_assign_decl_rhs(st, decl_rhs):
    st.decl_rhs = decl_rhs

@njit(types.void(RuleType, rhs_ptrs_type),cache=True)
def rule_assign_imper_rhs(st, imper_rhs):
    st.imper_rhs = imper_rhs




from numba import njit, types
from cre.utils import decode_idrec, used_bytes, NRTStatsEnabled
from cre.context import cre_context 
from cre.memset import MemSet, MemSetType 
from cre.transform.flattener import Flattener, flattener_update
from cre.fact import define_fact
import pytest

def flat_ms_vals(flat_ms):
    from cre.gval import gval
    t_ids = set([decode_idrec(x.idrec)[0] for x in flat_ms.get_facts(gval)])
    values = set([x.val for x in flat_ms.get_facts(gval)])
    return values


def test_flatten():
    with cre_context("test_flatten") as context:
        spec1 = {"A" : {"type" : "string", "visible" : True}, 
                 "B" : {"type" : "number", "visible" : False}}
        BOOP1 = define_fact("BOOP1", spec1)
        spec2 = {"inherit_from" : BOOP1,
                 "C" : {"type" : "number", "visible" : True}}
        BOOP2 = define_fact("BOOP2", spec2)
        spec3 = {"inherit_from" : BOOP2,
                 "D" : {"type" : "number", "visible" : True}}
        BOOP3 = define_fact("BOOP3", spec3)


        ms = MemSet()
        a = BOOP1("A", 1)
        b = BOOP1("B", 2)
        c = BOOP2("C", 3, 13)
        d = BOOP2("D", 4, 14)
        e = BOOP3("E", 5, 15, 105)
        f = BOOP3("F", 6, 16, 106)

        a_idrec = ms.declare(a)
        b_idrec = ms.declare(b)
        c_idrec = ms.declare(c)
        d_idrec = ms.declare(d)
        e_idrec = ms.declare(e)
        f_idrec = ms.declare(f)
        print("-------")

        fl = Flattener((BOOP1, BOOP2, BOOP3), ms, id_attr="A")
        
        out_ms = fl()
        values = flat_ms_vals(out_ms)
        print(out_ms)
        print(values)
        assert values == {"A", "B", "C", "D", "E" ,"F", 13., 14., 15., 16., 105.,106.}

        ms.retract(c)
        ms.retract(d)

        out_ms = fl()
        values = flat_ms_vals(out_ms)
        print(values)
        assert values == {"A", "B", "E" ,"F", 15., 16., 105.,106.}

        ms.modify(e, "D", 777)
        ms.modify(f, "A", "Z")

        out_ms = fl()
        values = flat_ms_vals(out_ms)
        print(values)
        assert values == {"A", "B", "E" ,"Z", 15., 16., 777., 106.}




def test_fl_mem_leaks():
    with NRTStatsEnabled:
        with cre_context("test_fl_mem_leaks"):
            spec ={ "A" : {"type" : "string", "visible" : True},
                    "B" : {"type" : "number", "visible" : True}
              }
            BOOP = define_fact("BOOP", spec)

            for k in range(5):
                ms = MemSet()
                for i in range(10):
                    ms.declare(BOOP(str(i),i))

                fl = Flattener([BOOP],in_memset=ms,id_attr="A")
                flat_ms = fl(ms)
                if(k <= 1):
                    init_bytes = used_bytes()

            assert used_bytes()-init_bytes == 0


        
        # print(fl._meminfo.refcount)
        # fl = None
        # print("<<", used_bytes()-init_bytes)
        # print(flat_ms._meminfo.refcount)
        # flat_ms = None
        # print("<<", used_bytes()-init_bytes)
        # print(ms._meminfo.refcount)
        # ms = None
        # print("<<", used_bytes()-init_bytes)


        # do_update(*args, **kwargs)

        

# with cre_context("flat") as context:

# with cre_context("flatten_10000"):
#     spec ={ "A" : {"type" : "string", "visible" : True},
#             "B" : {"type" : "number", "visible" : True}
#           }
#     BOOP = define_fact("BOOP", spec)
#     print("INIT SPEC", BOOP.spec)
#             # return (BOOP,), {}

#     @njit(cache=True)
#     def _b_dec_10000(ms):
#         for i in range(10000):
#             b = BOOP("HI",i)
#             ms.declare(b)

def setup_flatten():
    spec ={ "A" : {"type" : "string", "visible" : True},
            "B" : {"type" : "number", "visible" : True}
      }
    BOOP = define_fact("BOOP", spec)

    @njit(types.void(MemSetType), cache=True)
    def _b_dec_10000(ms):
        for i in range(10000):
            b = BOOP("HI",i)
            ms.declare(b)

    print("SPEC:", BOOP.spec)
    ms = MemSet()
    ms.declare(BOOP("HI",-1))
    fl = Flattener((BOOP,),in_memset=ms,id_attr="A",)
    fl.update()
    _b_dec_10000(ms)
        
    return (fl,ms), {}

def do_flatten(fl,ms):
    fl.update()
    return fl.out_memset

@pytest.mark.benchmark(group="transform")
def test_b_flatten_10000(benchmark):
    with cre_context("flatten_10000") as context:
        benchmark.pedantic(do_flatten,setup=setup_flatten, warmup_rounds=1, rounds=10)

if(__name__ == "__main__"):
    import faulthandler; faulthandler.enable()
    test_fl_mem_leaks()
    # test_flatten()
    # from cre.utils import PrintElapse
    # fl = setup_flatten()[0][0]
    # with PrintElapse("ABC"):
    #     fl.update()
    # fl = setup_flatten()[0][0]
    # with PrintElapse("ABC"):
    #     fl.update()


# ms.declare(BOOP1("zA", 1))
# ms.declare(BOOP1("zB", 2))
# ms.declare(BOOP2("zC", 3, 13))
# ms.declare(BOOP2("zD", 4, 14))
# ms.declare(BOOP3("zE", 5, 15, 105))
# ms.declare(BOOP3("zF", 6, 16, 106))

# flattener_update(fl)

# print(ms,out_ms)
# flattener_ctor((BOOP1, BOOP2, BOOP3))


# raise ValueError()



from numba import generated_jit, njit, i8, f8
from numba.types import unicode_type, FunctionType
from numba.core.errors import NumbaError, NumbaPerformanceWarning
from cre.func import CREFunc, set_func_arg, set_var_arg, reinitialize, CREFuncTypeClass, cre_func_copy
from cre.obj import _get_chr_mbrs_infos_from_attrs, _iter_mbr_infos
from cre.fact import define_fact
from cre.var import Var
from cre.utils import PrintElapse, _func_from_address, _cast_structref, NRTStatsEnabled, used_bytes
from cre.context import cre_context
import cre.type_conv
import pytest




def test_numerical():
    with NRTStatsEnabled:
        @CREFunc(signature=i8(i8,i8,i8,i8))
        def Add(a, b, c, d):
            return a + b + c +d

        a = Var(i8,'a')
        b = Var(i8,'b')
        c = Var(i8,'c')

        z = Add(a,b,c,c)
        assert z.depth == 1

        # print("<<", str(z))
        assert z(1,2,3) == 9
        assert str(z) == "Add(a, b, c, c)"
        q = z(c,c,c)
        # print("<<", q)
        assert(q(7)==28)
        assert str(q) == "Add(c, c, c, c)"

        assert Add(a,Add(a,b,c,a),c,c)(1,2,3) == 14
        assert Add(Add(Add(2,1,1,a),2,1,Add(2,1,1,b)),1,1,c)(1,2,3) == 19
        z = Add(Add(Add(2,1,1,a),2,1,Add(2,1,1,b)),1,1,c)
        assert str(z) == "Add(Add(Add(2, 1, 1, a), 2, 1, Add(2, 1, 1, b)), 1, 1, c)"
        assert z.depth == 3

        @CREFunc(signature=f8(f8,f8),
                shorthand='{0}+{1}')
        def Add(a, b):
            return a + b

        assert Add(9.0, 7.0) == 16.0

        for i in range(2):
            a = Var(f8,'a')
            Incr = Add(Var(f8,'a'),1.0)
            assert Incr(3.0) == 4.0
            assert str(Incr) == "a+1.0"

            if(i == 0):
                init_bytes = used_bytes()
            else:
                assert used_bytes() == init_bytes



def test_string():
    with NRTStatsEnabled:
        @CREFunc(signature=unicode_type(unicode_type,unicode_type),
                shorthand='{0}+{1}')
        def Concat(a, b):
            return a + b
        
        a = Var(unicode_type,'a')
        b = Var(unicode_type,'b')
        c = Var(unicode_type,'c')

        for i in range(2):
            z = Concat("|", Concat(Concat(b,c),"|" ))

            assert z("X","Y") == "|XY|"
            assert str(z) == "'|'+((b+c)+'|')"
            assert z.depth == 3

            if(i == 0):
                init_bytes = used_bytes()
            else:
                assert used_bytes() == init_bytes



def test_obj():
    with NRTStatsEnabled:
        with cre_context("test_obj"):
            BOOP = define_fact("BOOP", {"A" :unicode_type, "B" :i8})

            @CREFunc(signature=unicode_type(unicode_type,unicode_type),
                    shorthand='{0}+{1}')
            def Concat(a, b):
                return a + b

            @CREFunc(signature=BOOP(BOOP,BOOP),
                    shorthand='{0}+{1}')
            def CatBOOPs(a, b):
                return BOOP(a.A + b.A, a.B + b.B)

            a = Var(BOOP,'a')
            b = Var(BOOP,'b')
            c = Var(BOOP,'c')
            
            for i in range(2):
                ba, bb = BOOP("A",1), BOOP("B",2)

                z = Concat(a.A, a.A)
                print(z(ba), str(z))

                assert z(ba) == "AA"
                assert str(z) == "a.A+a.A"

                z = Concat(a.A, b.A)

                assert z(ba,bb) == "AB"
                assert str(z) == "a.A+b.A"

                z = Concat(Concat("|", Concat(a.A, b.A)),"|")
                assert z(ba,bb) == "|AB|"
                

                zboop = CatBOOPs(a,b)

                assert zboop(ba,bb) == BOOP(A="AB", B=3)
                assert str(zboop) == "a+b"

                # Make sure "head_var_ptrs" actually grabs vars w/ derefs
                assert not any(z.base_var_ptrs == z.head_var_ptrs)

                z = None
                zboop = None
                if(i == 0):
                    init_bytes = used_bytes()
                else:
                    assert used_bytes() == init_bytes
                    # print(used_bytes(), init_bytes)

def test_mixed_types():
    from cre.default_funcs import Identity
    @CREFunc(signature=i8(i8,i8),
            shorthand='{0}+{1}')
    def Add(a, b):
        return a + b

    @CREFunc(signature=unicode_type(unicode_type,unicode_type),
                shorthand='{0}+{1}')
    def Concat(a, b):
        return a + b

    @CREFunc(signature=unicode_type(i8),
            shorthand='str({0})')
    def ToStr(a):
        return str(a)

    @CREFunc(signature=i8(unicode_type),
            shorthand='int({0})')
    def ToInt(a):
        return float(a)

    a, b, c = Var(i8,'a'), Var(i8,'b'), Var(i8,'c')

    z = Concat(Concat(ToStr(a),ToStr(b)),ToStr(c))
    assert z(1,2,3) == "123"
    # print("ret_val:", ret_val)
    q = ToInt(z)
    assert q(1,2,3) == 123    

    z = ToStr(Add(Add(a,b),c))
    assert z(1,2,3) == "6"
    




new_f_type = CREFuncTypeClass(f8,(f8,f8),is_composed=True,name="Composed")
@njit(cache=True)
def compose_hardcoded(f, g):
    _f = cre_func_copy(f)
    g1 = cre_func_copy(g)
    g2 = cre_func_copy(g)
    set_var_arg(g1, 0, Var(f8,"x"))
    reinitialize(g1)
    set_var_arg(g2, 0, Var(f8,"y"))
    reinitialize(g2)
    set_func_arg(_f, 0, g1)
    set_func_arg(_f, 1, g2)
    reinitialize(_f)
    return _cast_structref(new_f_type, _f)

@njit(cache=True)
def compose_overloaded(f, g):
    
    x,y = Var(f8,"x"), Var(f8,"y")
    h = f(g(x),g(y))
    return h


def test_njit_compose():    
    @CREFunc(signature=f8(f8,f8), shorthand="{0}*{1}")
    def Multiply(a, b):
        return a * b

    @CREFunc(signature=f8(f8), shorthand="{0}+1")
    def Increment(a):
        return a + 1

    f = compose_hardcoded(Multiply, Increment)

    assert f(1,2) == 6
    assert str(f) == "(x+1)*(y+1)"

    f = compose_overloaded(Multiply, Increment)

    assert f(1,2) == 6
    assert str(f) == "(x+1)*(y+1)"

from cre.func import cre_func_deep_copy_generic
def test_no_mutate_on_compose():
    a,b,c = Var(f8,'a'), Var(f8,'b'), Var(f8,'c')
    c0 = a + b + b
    c0_ = cre_func_deep_copy_generic(c0)
    print("c0_", c0_)
    s0 = str(c0) 
    print(c0)
    c1 = c0(a,c+b)
    s1 = str(c0)
    print(c1)
    print(c0)
    assert str(c0) == s0

def test_compose_deref_bases():
    with cre_context("test_compose_deref_bases"):
        BOOP = define_fact("BOOP", {"A" :unicode_type, "B" : f8})

        @CREFunc(signature=f8(f8,f8), shorthand="{0}*{1}")
        def Multiply(a, b):
            return a * b

        a,b = Var(BOOP,'a'), Var(BOOP,'b')

        c0 = Multiply(a.B, Multiply(a.B, b.B))

        a0, a1 =BOOP("A",7),BOOP("B",2) 
        print(c0(a0,a1))
        assert c0(BOOP("A",7),BOOP("B",2)) == 98

        print(c0)

        x,y = Var(BOOP,'x'), Var(BOOP,'y')

        c1 = c0(x,y)

        assert c1(BOOP("A",7),BOOP("B",2)) == 98

        print(c1)




def test_commutes():
    @CREFunc(signature=f8(f8,f8,f8),
            shorthand="{0}*{1}*{2}", commutes=True)
    def Add3(a, b, c):
        return a + b + c

    print(Add3.commutes)
    assert Add3.commutes == [[0,1,2]]
    print(Add3.right_commutes)
    rc_str = str(Add3.right_commutes)
    assert "2: array([0, 1]" in rc_str
    assert "1: array([0]" in rc_str

    with pytest.raises(AssertionError):
        @CREFunc(signature=f8(f8,f8, unicode_type),commutes=[[0,1,2]])
        def Floop(a,b,c):
            return a + b

def test_not_jittable():
    from cre.default_funcs import Add

    with pytest.warns(NumbaPerformanceWarning):
        @CREFunc(signature=f8(f8))
        def Map(a):
            l = []
            d = {1.0: 10.0, 2.0: 20.0}
            for i in range(int(a)):
                l.append(str(i+1))
                l.append(float(i+1))

            return d[l[-1]]


    assert Map(1.0)==10.0
    # assert Map.call(1.0)==10.0
    # assert Map.check(1.0)==1


    op = Add(Map(Var(float,'x')),Map(Var(float,'y')))
    assert op(1,2)==30.0


def test_returns_object():
    with cre_context('test_returns_object'):
        spec = {"A" : "string", "B" : "number"}
        BOOP = define_fact("BOOP", spec)


        @CREFunc(signature=BOOP(BOOP))
        def RetBOOP(a):
            return BOOP(a.A, a.B+9.0)
            
        assert RetBOOP(BOOP("A",1.0)).B == 10.0

        with pytest.warns(NumbaPerformanceWarning):        
            @CREFunc(signature=BOOP(BOOP))
            def BOOPMap(a):
                l = []
                d = {1.0: 10.0, 2.0: 20.0}
                for i in range(int(a.B)):
                    l.append(str(i+1))
                    l.append(float(i+1))

                return BOOP("A", d[l[-1]])

        assert BOOPMap(BOOP("A",1.0)).B == 10.0

def test_constant():
    @CREFunc(signature=unicode_type(), shorthand="'X'")
    def X():
        return "X"
    assert str(X) == "'X'"
    assert X() == "X"


def test_var_cmp_overloads():
    x, y = Var(f8,'x'), Var(f8,'y')

    #--beta cmp--
    op = x < y
    assert str(op) == "x < y"
    op = x <= y
    assert str(op) == "x <= y"
    op = x > y
    assert str(op) == "x > y"
    op = x >= y
    assert str(op) == "x >= y"
    op = x == y
    assert str(op) == "x == y"

    #--left cmp--
    op = 1 < y
    assert str(op) == "y > 1"
    op = 1 <= y
    assert str(op) == "y >= 1"
    op = 1 > y
    assert str(op) == "y < 1"
    op = 1 >= y
    assert str(op) == "y <= 1"
    op = 1 == y
    assert str(op) == "y == 1"

    #--right cmp--
    op = x < 1
    assert str(op) == "x < 1"
    op = x <= 1
    assert str(op) == "x <= 1"
    op = x > 1
    assert str(op) == "x > 1"
    op = x >= 1
    assert str(op) == "x >= 1"
    op = x == 1
    assert str(op) == "x == 1"

def test_var_arith_overloads():
    x, y = Var(f8,'x'), Var(f8,'y')

    #--beta arith--
    op = x + y
    assert str(op) == "x + y"
    op = x - y
    assert str(op) == "x - y"
    op = x * y
    assert str(op) == "x * y"
    op = x / y
    assert str(op) == "x / y"
    op = x // y
    assert str(op) == "x // y"
    op = x ** y
    assert str(op) == "x ** y"

    #--left arith--
    op = 1 + y
    assert str(op) == "1 + y"
    op = 1 - y
    assert str(op) == "1 - y"
    op = 1 * y
    assert str(op) == "1 * y"
    op = 1 / y
    assert str(op) == "1 / y"
    op = 1 // y
    assert str(op) == "1 // y"
    op = 1 ** y
    assert str(op) == "1 ** y"

    #--right arith--
    op = x + 1
    assert str(op) == "x + 1"
    op = x - 1
    assert str(op) == "x - 1"
    op = x * 1
    assert str(op) == "x * 1"
    op = x / 1
    assert str(op) == "x / 1"
    op = x // 1
    assert str(op) == "x // 1"
    op = x ** 1
    assert str(op) == "x ** 1"



def test_op_cmp_overloads():
    x, y, z = Var(f8,'x'), Var(f8,'y'), Var(f8,'z')

    #--beta cmp--
    op = (x + z) < (y + z)
    assert str(op) == "(x + z) < (y + z)"
    op = (x + z) <= (y + z)
    assert str(op) == "(x + z) <= (y + z)"
    op = (x + z) > (y + z)
    assert str(op) == "(x + z) > (y + z)"
    op = (x + z) >= (y + z)
    assert str(op) == "(x + z) >= (y + z)"
    op = (x + z) == (y + z)
    assert str(op) == "(x + z) == (y + z)"

    # #--left cmp--
    op = 1 < (y + z)
    assert str(op) == "(y + z) > 1"
    op = 1 <= (y + z)
    assert str(op) == "(y + z) >= 1"
    op = 1 > (y + z)
    assert str(op) == "(y + z) < 1"
    op = 1 >= (y + z)
    assert str(op) == "(y + z) <= 1"
    op = 1 == (y + z)
    assert str(op) == "(y + z) == 1"

    # #--right cmp--
    op = (x + z) < 1
    assert str(op) == "(x + z) < 1"
    op = (x + z) <= 1
    assert str(op) == "(x + z) <= 1"
    op = (x + z) > 1
    assert str(op) == "(x + z) > 1"
    op = (x + z) >= 1
    assert str(op) == "(x + z) >= 1"
    op = (x + z) == 1
    assert str(op) == "(x + z) == 1"

def test_op_arith_overloads ():
    x, y, z = Var(f8,'x'), Var(f8,'y'), Var(f8,'z')

    #--beta cmp--
    op = (x + z) + (y + z)
    assert str(op) == "(x + z) + (y + z)"
    op = (x + z) - (y + z)
    assert str(op) == "(x + z) - (y + z)"
    op = (x + z) * (y + z)
    assert str(op) == "(x + z) * (y + z)"
    op = (x + z) / (y + z)
    assert str(op) == "(x + z) / (y + z)"
    op = (x + z) // (y + z)
    assert str(op) == "(x + z) // (y + z)"
    op = (x + z) ** (y + z)
    assert str(op) == "(x + z) ** (y + z)"

    # #--left cmp--
    op = 1 + (y + z)
    assert str(op) == "1 + (y + z)"
    op = 1 - (y + z)
    assert str(op) == "1 - (y + z)"
    op = 1 * (y + z)
    assert str(op) == "1 * (y + z)"
    op = 1 / (y + z)
    assert str(op) == "1 / (y + z)"
    op = 1 // (y + z)
    assert str(op) == "1 // (y + z)"
    op = 1 ** (y + z)
    assert str(op) == "1 ** (y + z)"

    # #--right cmp--
    op = (x + z) + 1
    assert str(op) == "(x + z) + 1"
    op = (x + z) - 1
    assert str(op) == "(x + z) - 1"
    op = (x + z) * 1
    assert str(op) == "(x + z) * 1"
    op = (x + z) / 1
    assert str(op) == "(x + z) / 1"
    op = (x + z) // 1
    assert str(op) == "(x + z) // 1"
    op = (x + z) ** 1
    assert str(op) == "(x + z) ** 1"

def test_ptr_ops():
    from cre.default_funcs import ObjIsNone, ObjEquals
    with cre_context("test_ptr_ops"):
        BOOP = define_fact("BOOP", {"val" : f8,"nxt" : "BOOP"})

        a,b,c = Var(BOOP,"a"), Var(BOOP,"b"), Var(BOOP,"c")    

        l1 = ObjIsNone(a.nxt)
        l2 = ObjEquals(a.nxt, b.nxt)

        assert str(l2) == 'a.nxt == b.nxt'
        # print(repr(l2))
        # assert repr(l2) == 'ObjEquals([a:CREObjType].nxt, [b:CREObjType].nxt)'

        l3 = ObjEquals(a, a.nxt)
        assert str(l3) == 'a == a.nxt'

        x0 = BOOP(1.0)
        x1 = BOOP(1.0)
        y0 = BOOP(2.0,x0)
        assert ObjEquals(x0,x0)
        assert not ObjEquals(x0,y0)
        assert not ObjEquals(x0,x1)
        assert ObjEquals(a, b.nxt)(x0, y0)

def test_bad_compose():
    a, b = Var(f8,'a'), Var(f8,'b')
    x, y = Var(unicode_type,'x'), Var(unicode_type,'y')

    # Var case
    q = a + b
    with pytest.raises(TypeError):
        q(x,y)

    # CREFunc case
    q_s = x + y
    with pytest.raises(TypeError):
        q_s(q,q)

    # Constant case
    with pytest.raises(TypeError):
        q("A","B")

        
def test_minimal_str():
    from cre.default_funcs import CastFloat, CastStr
    with cre_context("test_minimal_str"):
        BOOP = define_fact("BOOP", 
            {"id": unicode_type, "val" : f8}
        )

        a,b = Var(BOOP,"a"), Var(BOOP,"b")

        c = CastStr(CastFloat(a.id) + (b.val))

        assert str(c) == "str(float(a.id) + b.val)"
        assert c.minimal_str(ignore_funcs=[CastFloat, CastStr]) == "a + b"

        c = CastStr(CastFloat(a.id) + (b.val) + 100)        

        assert str(c) == "str((float(a.id) + b.val) + 100)"
        assert c.minimal_str(ignore_funcs=[CastFloat, CastStr]) == "(a + b) + 100"


# ---------------------------------------------------
# : Performance Benchmarks

def setup_call():
    @CREFunc(signature=i8(i8,i8))
    def Add(a, b):
        return a + b

    @CREFunc(signature=i8(i8,i8,i8))
    def Add3(a, b, c):
        return a + b + c

    a = Var(i8,'a')
    b = Var(i8,'b')
    comp = Add3(0,a,b)
    return (Add, comp), {}

N = 100

@njit(cache=True)
def apply_100x100(op):
    z = 0
    for i in range(N):
        for j in range(N):
            z += op(i, j)
    return z

@pytest.mark.benchmark(group="cre_func")
def test_b_call_inlined_uncomposed_100x100(benchmark):
    benchmark.pedantic(lambda op,comp: apply_100x100(op),
        setup=setup_call, warmup_rounds=1, rounds=10)

@pytest.mark.benchmark(group="cre_func")
def test_b_call_composed_100x100(benchmark):
    benchmark.pedantic(lambda op,comp: apply_100x100(comp),
        setup=setup_call, warmup_rounds=1, rounds=10)

i8_ft = FunctionType(i8(i8,i8))
@njit(cache=True)
def call_heads_100x100(op):
    z = 0
    for i in range(N):
        for j in range(N):
            f = _func_from_address(i8_ft,op.call_heads_addr)
            z += f(i,j)   
    return z    

@pytest.mark.benchmark(group="cre_func")
def test_b_call_dynamic_100x100(benchmark):
    benchmark.pedantic(lambda op,comp: call_heads_100x100(op),
        setup=setup_call, warmup_rounds=1, rounds=10)


def setup_compose():
    @CREFunc(signature=i8(i8,i8,i8))
    def Add3(a, b, c):
        return a + b + c

    a = Var(i8,'a')
    b = Var(i8,'b')
    return (Add3, a, b), {}

@njit(cache=True)
def compose(Add3, a, b):
    for i in range(100):
        Add3(i,a,b)

@pytest.mark.benchmark(group="cre_func")
def test_compose_py_100(benchmark):
    benchmark.pedantic(compose.py_func,
        setup=setup_compose, warmup_rounds=1, rounds=10)

@pytest.mark.benchmark(group="cre_func")
def test_compose_nb_100(benchmark):
    benchmark.pedantic(compose,
        setup=setup_compose, warmup_rounds=1, rounds=10)



# class ShouldRaise():
#     def __enter__(self):



if __name__ == "__main__":
    import faulthandler; faulthandler.enable()
    import sys

    # @CREFunc(signature=i8(i8,i8,i8))
    # def Add3(a, b, c):
    #     return a + b + c

    # from cre.utils import _tuple_getitem
    # @njit
    # def bar():
    #     return _tuple_getitem((1,"2",False), 1)

    # print(bar())
    # # raise ValueError()

    
    # @njit
    # def foo(f,a,b):
    #     f(7, a, b)

    # a = Var(i8,'a')
    # b = Var(i8,'b')
    # # foo(Add3,a,b)
    # with PrintElapse("Compose 100"):
    #     for i in range(100):
    #         Add3(i,a,b)

    

    # test_commutes()
    # test_no_mutate_on_compose()
    # test_numerical()
    # test_string()
    # test_obj()
    # test_mixed_types()
    # test_njit_compose()
    # test_compose_deref_bases()
    test_not_jittable()
    # test_returns_object()
    # test_var_cmp_overloads()
    # test_var_arith_overloads()
    # test_op_cmp_overloads()
    # test_op_arith_overloads()
    # test_ptr_ops()
    # test_constant()
    # test_bad_compose()
    # test_minimal_str()

    sys.exit()
    # @njit(f8(f8,f8),cache=True)
    with PrintElapse("DEFINE DIVIDE"):
        @CREFunc(signature=f8(f8,f8),
                shorthand='{0}/{1}')
        def Divide(a, b):
            if(a == 0):
                raise ValueError("Bad a")
            else:
                return a / b

    with PrintElapse("DEFINE QDIVIDE"):
        @CREFunc(signature=f8(f8,f8),
                shorthand='{0}/{1}')
        def ZDivide(a, b):
            if(a == 0):
                raise ValueError("Bad a")
            else:
                return a / b

    sys.exit() # Stuff below still has issues


    with pytest.raises(ValueError):
        Divide(0,2)
    
    with pytest.raises(ValueError):
        Divide(0,2)    

    print("::")
    # with pytest.raises(ZeroDivisionError):
    print(Divide(1,0))

    @njit(cache=True)
    def boop(op,a,b):
        op(a,b)

    boop(Divide, 0,2)
    # boop(Divide, 1,0)



    f8_ft = FunctionType(f8(f8,f8))
    @njit(cache=True)
    def boop(cf,a,b):
        fn = _func_from_address(f8_ft, cf.call_heads_addr)
        fn(a,b)

    boop(Divide, 0,2)
    # boop(Divide, 1,0)

    # print(list(Divide.overloads.values())[0].type_annotation.__dict__)
    # print(list(Divide.overloads.values())[0].entry_point)

    











import numpy as np
import numba
from numba import f8, i8, njit
from numba.typed import List, Dict
from numba.types import ListType, DictType, unicode_type
from cre.memset import MemSet
from cre.var import Var
from cre.transform.flattener import Flattener
from cre.transform.feature_applier import FeatureApplier
from cre.transform.relative_encoder import _check_needs_rebuild, RelativeEncoder, get_relational_fact_attrs, next_adjacent
from cre.utils import PrintElapse, deref_info_type, used_bytes, NRTStatsEnabled
from cre.fact import define_fact
from cre.default_funcs import Equals
from cre.context import cre_context
import cre
from numba.core.runtime.nrt import rtsys
import gc
import pytest
# v1 = Var(Container).children[0]
# v2 = Var(Container).children[1]
# print(v1, v1.deref_infos, v1.deref_infos)
# print(v2, v2.deref_infos, v2.deref_infos)




with cre_context("test_relative_encoder"):
    eq_f8 = Equals(f8, f8)
    eq_str = Equals(unicode_type, unicode_type)


def def_fact_types():
    Component = define_fact("Component", {
        "id" : unicode_type,
        "value" : {"type" : unicode_type, "visible" : True},
        "above" : "Component", "below" : "Component",
        "to_left": "Component", "to_right" : "Component",
        "parents" : "List(Component)"
    })

    Container = define_fact("Container", {
        "inherit_from" : "Component",
        "children" : "List(Component)"
    })

    TestLL = define_fact("TestLL",{
        "id": "string",
        "value" : {"type": "string", "visible" : True},
        "nxt" : "TestLL",
        "prev" : "TestLL",
    })
    return Component, Container, TestLL



def setup_encoder_w_heir_state():
    with cre_context("test_relative_encoder"):
        Component, Container, TestLL = def_fact_types()
        ### Make Structure ### 
        #     p3
        #     p2
        #     p1
        #  [a,b,c]

        a = Component(id="A",value="a")
        b = Component(id="B",value="b")
        c = Component(id="C",value="c")
        a.to_right = b
        b.to_right = c
        b.to_left = a
        c.to_left = b

        p1 = Container(id="P1", children=List([a,b,c]))
        a.parents = List([p1])
        b.parents = List([p1])
        c.parents = List([p1])

        p2 = Container(id="P2", children=List([p1]))
        p1.parents = List([p2])

        p3 = Container(id="P3", children=List([p2]))
        p2.parents = List([p3])

        ms = MemSet()
        ms.declare(a)
        ms.declare(b)
        ms.declare(c)
        ms.declare(p1)
        ms.declare(p2)
        ms.declare(p3)

        fl = Flattener((Component,Container,), ms, id_attr="id")
        flat_ms = fl()
        fa = FeatureApplier([eq_f8,eq_str], flat_ms)
        feat_ms = fa()

        re = RelativeEncoder((Component,Container),ms)
        vs = [Var(Container,'p1'),Var(Container,'p2'),Var(Container,'p3')]
        return (ms,feat_ms,fl,fa,re,[p1,p2,p3], vs), {}



def test_relative_encoder():
    with cre_context("test_relative_encoder"):
        import faulthandler; faulthandler.enable()

        (ms,feat_ms,fl,fa,re,ps,vs),_ = setup_encoder_w_heir_state()
        p1, p2, p3  = ps[0],ps[1],ps[2]
        vp1,vp2,vp2 = vs[0],vs[1],vs[2]

        print(ms)
        print(feat_ms)
        re.update()

        with PrintElapse("encode_relative"):
            rel_ms = re.encode_relative_to(feat_ms,[p1], [vp1])
        print(rel_ms)

        with PrintElapse("encode_relative"):
            rel_ms = re.encode_relative_to(feat_ms,[p1,p2,p3], [vp1,vp2,vp2])
        print(rel_ms)
    


def setup_update():
    with cre_context("test_relative_encoder"):
        Component, Container, TestLL = def_fact_types()

        ms = MemSet()
        
        fl = Flattener((TestLL,), in_memset=ms, id_attr="id",)
        end = first = TestLL(id="q",value="1")
        
        flat_ms = fl()
        fa = FeatureApplier([eq_f8,eq_str], flat_ms)
        feat_ms = fa()

        re = RelativeEncoder((TestLL,), ms, id_attr="id")
        re.update()

        ms.declare(end)

        for i in range(100):
            # erc0 = end._meminfo.refcount
            new_end = TestLL(str(i),str(i),end)
            # erc1, nrc1 = end._meminfo.refcount, new_end._meminfo.refcount 
            # print(f'{erc0}->{erc1}', f'{nrc1}')
            ms.modify(end, "prev",new_end)
            ms.declare(new_end)
            # erc2, nrc2 = end._meminfo.refcount, new_end._meminfo.refcount 
            # print(f'{erc1}->{erc2}', f'{nrc1}->{nrc2}')
            end = new_end

        flat_ms = fl()
        feat_ms = fa()
        feat_ms, fa = None, None

    return (re,ms,end), {}

def setup_update_plus_1():
    with cre_context("test_relative_encoder"):
        Component, Container, TestLL = def_fact_types()
        (re,ms,end),_ = setup_update()
        re.update()
        new_end = TestLL("plus1","plus1",end)
        ms.modify(end, "prev",new_end)
        ms.declare(new_end)
        
        return (re,ms,end), {}    



def do_update(re,ms,end):
    re.update()

def do_encode_rel(ms,feat_ms,fl,fa,re,ps,vs):
    rel_ms = re.encode_relative_to(feat_ms, ps,vs)


# NOTE: Requires revisiting. Definitely partial leak here.
#   The relative encoder seems to be aquiring a reference
#   to each of the facts in the memset at instantiation.
def test_re_mem_leaks():
    with NRTStatsEnabled:
        with cre_context("test_re_mem_leaks"):
            for i in range(5):
                args, kwargs = setup_update()
                (re,ms,end) = args
                # first = ms.get_facts()[0]

                # print(re._meminfo.refcount)
                # print(ms._meminfo.refcount)
                # print("first", first._meminfo.refcount)
                # print("end", end._meminfo.refcount)
                ms.clear_refs()
                
                # print("Free first", first._meminfo.refcount)
                # print("Free end", end._meminfo.refcount)

                ms =None; gc.collect()

                # print("msnull first", first._meminfo.refcount)
                # print("msnull end", end._meminfo.refcount)

                # NOTE: really ought to be fine when == 0
                if(i <= 1):
                    init_bytes = used_bytes()
                else:
                    print("<<", used_bytes()-init_bytes)

            # Commenting out for now until larger memleak issue is resolved for facts
            assert used_bytes()-init_bytes == 0


def test_downcasted_heads():
    with cre_context("test_downcasted_heads"):
        Component = define_fact("Component", {
            "id" : str,
            "x" : {"type" : float, "visible" : False},
            "y" : {"type" : float, "visible" : False},
            "width" : {"type" : float, "visible" : False},
            "height" : {"type" : float, "visible" : False},
            "above" : "Component", 
            "below" : "Component",
            "left": "Component", 
            "right" : "Component",
            "parents" : "List(Component)"
        })

        TextField = define_fact("TextField", {
            "inherit_from" : "Component",
            "value" : {"type" : str, "visible" : True, "semantic" : True},
            "locked" : {"type" : bool, "visible" : True},
        })

        Button = define_fact("Button", {
            "inherit_from" : "Component",
        })

        Container = define_fact("Container", {
            "inherit_from" : "Component",
            "children" : "List(Component)"
        })

        a = TextField(id="A",value="a")
        b = TextField(id="B",value="b")
        c = TextField(id="C",value="c")
        a.right = b
        b.right = c
        b.left = a
        c.left = b

        p1 = Container(id="P1", children=List([a,b,c]))
        a.parents = List([p1])
        b.parents = List([p1])
        c.parents = List([p1])

        p2 = Container(id="P2", children=List([p1]))
        p1.parents = List([p2])

        p3 = Container(id="P3", children=List([p2]))
        p2.parents = List([p3])

        ms = MemSet()
        ms.declare(a)
        ms.declare(b)
        ms.declare(c)
        ms.declare(p1)
        ms.declare(p2)
        ms.declare(p3)

        fl = Flattener((TextField,Container,), ms, id_attr="id")
        flat_ms = fl()
        fa = FeatureApplier([eq_f8,eq_str], flat_ms)
        feat_ms = fa()

        print(feat_ms)

        re = RelativeEncoder((TextField,Container),ms)
        vs = [Var(Container,'p1'),Var(Container,'p2'),Var(Container,'p3')]

        vp1,vp2,vp2 = vs[0],vs[1],vs[2]


        print(ms)
        re.update()

        with PrintElapse("encode_relative"):
            rel_ms = re.encode_relative_to(feat_ms,[p1], [vp1])
        print(rel_ms)

        with PrintElapse("encode_relative"):
            rel_ms = re.encode_relative_to(feat_ms,[p1,p2,p3], [vp1,vp2,vp2])
        print(rel_ms)

        with PrintElapse("encode_relative"):
            rel_ms = re.encode_relative_to(feat_ms,[a,c], [Var(TextField,'a'),Var(TextField,'c')])
        print(rel_ms)

    
    
        

@pytest.mark.benchmark(group="transform")
def test_b_rel_enc_update_100x100(benchmark):
    with cre_context("test_relative_encoder"):
        benchmark.pedantic(do_update,setup=setup_update, warmup_rounds=1, rounds=10)

@pytest.mark.benchmark(group="transform")
def test_b_rel_enc_100x100_update_plus_1(benchmark):
    with cre_context("test_relative_encoder"):
        benchmark.pedantic(do_update,setup=setup_update_plus_1, warmup_rounds=1, rounds=10)

@pytest.mark.benchmark(group="transform")
def test_b_rel_enc_100x100_encode(benchmark):
    with cre_context("test_relative_encoder"):
        benchmark.pedantic(do_encode_rel,setup=setup_encoder_w_heir_state, warmup_rounds=1, rounds=10)


if __name__ == "__main__":
    import faulthandler; faulthandler.enable()
    test_relative_encoder()
    # test_re_mem_leaks()

    # test_downcasted_heads()
    # test_relative_encoder()

    # np.set_printoptions(linewidth=1000)
    # test_relative_encoder()
    # (re,ms,_), _ = setup_update()
    # with PrintElapse("Elapse"):
    #     re.update()
    # with PrintElapse("Elapse 2"):
    #     re.update()

    # (re,ms,_), _ = setup_update_plus_1()
    # with PrintElapse("Elapse_plus 1"):
    #     re.update()
